package jia;

import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.Atom;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.Term;

import java.util.ArrayList;
import java.util.Collections;
import java.util.logging.Level;

import prop.PropertiesReader;

import arch.CowboyArch;
import arch.LocalWorldModel;
import env.WorldModel;

public class astar extends DefaultInternalAction {
	
	public LocalWorldModel model = null;
	
    Node[][] nodes = null;
	
	public astar() {
	}
	
    @Override
    public Object execute(TransitionSystem ts, Unifier un, Term[] args) throws Exception {
        try {
        	
            int myX = (int)((NumberTerm)args[0]).solve();
			int myY = (int)((NumberTerm)args[1]).solve();
			int targetX = (int)((NumberTerm)args[2]).solve();
			int targetY = (int)((NumberTerm)args[3]).solve();
			
			String sAction = getActionFromAstar(astar(ts, myX, myY, targetX, targetY));

			return un.unifies(args[4], new Atom(sAction));
            
        } catch (Throwable e) {
            ts.getLogger().log(Level.SEVERE, "astar error: "+e, e);
        }
        return false;
    } // execute
    
    public ArrayList<Node> astar(TransitionSystem ts, int startX, int startY, int targetX, int targetY) {
    	return astar(ts, startX, startY, targetX, targetY, true);
    }
    
    public boolean isWalkable(Node n) {
    	if (n.content == WorldModel.OBSTACLE ||
    			n.content == WorldModel.FENCE ||
    			n.content == WorldModel.SWITCH)
    		return false;
    	
    	return true;
    }
    
    public ArrayList<Node> astar(TransitionSystem ts, int startX, int startY, int targetX, int targetY, boolean cowPunishment) {
    	CowboyArch arch = (CowboyArch)ts.getUserAgArch();
        model = arch.getModel();
    	
        ArrayList<Node> path = new ArrayList<Node>();
        
		// create local world
		nodes = new Node[model.getWidth()][model.getHeight()];
		
		for (int i = 0; i < model.getWidth(); i++) {
			for (int j = 0; j < model.getHeight(); j++) {
				Node n = new Node(i, j);
				n.content = getContent(n.x, n.y);
				nodes[i][j] = n;
			}
		}
		
	
		for (int i = 1; i < model.getWidth()-1; i++) {
			for (int j = 1; j < model.getHeight()-1; j++) {
				// horizontal filter
				if (!isWalkable(nodes[i-1][j]) && isWalkable(nodes[i][j]) && !isWalkable(nodes[i+1][j])){
					nodes[i][j].content = WorldModel.OBSTACLE;
				}
				
				// vertical filter
				if (!isWalkable(nodes[i][j-1]) && isWalkable(nodes[i][j]) && !isWalkable(nodes[i][j+1])){
					nodes[i][j].content = WorldModel.OBSTACLE;
				}
			}
		}
		
		
		
		// astar
		ArrayList<Node> openSet = new ArrayList<Node>();
		ArrayList<Node> closedSet = new ArrayList<Node>();
		
		// create initial node
		Node startNode = nodes[startX][startY];
		Node goalNode = nodes[targetX][targetY];
		
		openSet.add(startNode);
		
		calculatePunishment(cowPunishment);
		
		int count = 0;
		
		// iterate through openset
		while (openSet.size() > 0) {
			count++;
			Collections.sort(openSet);
			
			Node currentNode = openSet.remove(0);
			closedSet.add(currentNode);
			
			// check for goal
			if (currentNode == goalNode) {
				path.add(goalNode);
				Node selectedNode = currentNode;
				while (selectedNode.from != startNode) {
					selectedNode = selectedNode.from;
					path.add(selectedNode);
				}
				path.add(startNode);
				break;
			}
			
			ArrayList<Node> neighbours = getSurroundingNodes(ts, currentNode);
			for (Node n : neighbours) {
				if (closedSet.contains(n))
					continue;
				
				if (n.content == WorldModel.OBSTACLE
    	    			|| n.content == WorldModel.SWITCH)
					continue;
				
				double tempDist = currentNode.walkDistance + 1 + n.punishment;
				boolean tempDistIsBest = false;
				
				if (openSet.contains(n) == false) {
					openSet.add(n);
					double heuristicPower = PropertiesReader.getInstance().getDoubleProperty("heuristic_power");
					n.heuristic = Math.pow(n.distance(goalNode), heuristicPower);
					tempDistIsBest = true;
				}
				else if (tempDist < n.walkDistance)
					tempDistIsBest = true;
				
				if (tempDistIsBest) {
					n.from = currentNode;
					n.walkDistance = tempDist;
					n.totalDitance = n.walkDistance + n.heuristic;
				}
			} // for neighbours
		} //while
		
		Collections.reverse(path);
		return path;
    }
    
    public String getActionFromAstar(ArrayList<Node> path) {
    	if (path.size() < 2)
    		return "skip";
    	
    	return getMove(path.get(0), path.get(1));
    }
    
    public ArrayList<Node> getSurroundingNodes(TransitionSystem ts, Node n) {
    	ArrayList<Node> surroundingNodes = new ArrayList<Node>();
    	
    	for (int i = n.x-1; i <= n.x+1; i++) {
    		for (int j = n.y-1; j <= n.y+1; j++) {
    			if (i >= 0 && i < nodes.length && j >= 0 && j < nodes[0].length) {
    				surroundingNodes.add(nodes[i][j]);
    			}
    		}
    	}
    	return surroundingNodes;
    }
    
    public void calculatePunishment(boolean cowPunishment) {
    	double COW_RADIUS = PropertiesReader.getInstance().getDoubleProperty("cow_radius");
    	double COW_VALUE = PropertiesReader.getInstance().getDoubleProperty("cow_value");
    	double OBSTACLE_RADIUS = PropertiesReader.getInstance().getDoubleProperty("obstacle_radius");
    	double OBSTACLE_VALUE = PropertiesReader.getInstance().getDoubleProperty("obstacle_value");
    	double CLEAN_VALUE = PropertiesReader.getInstance().getDoubleProperty("clean_value");
    	double CORRAL_VALUE = PropertiesReader.getInstance().getDoubleProperty("corral_value");
    	double AGENT_VALUE = PropertiesReader.getInstance().getDoubleProperty("agent_value");
    	double FENCE_VALUE = PropertiesReader.getInstance().getDoubleProperty("fence_value");
    	double UNEXPLORED_VALUE = PropertiesReader.getInstance().getDoubleProperty("unexplored_value");
    	
    	for (int x = 0; x < nodes.length; x++) {
    		for (int y = 0; y < nodes[0].length; y++) {
    			
    			if (nodes[x][y].content == WorldModel.COW && cowPunishment) {  		    	
    		    	for (int i = nodes[x][y].x-(int)COW_RADIUS; i <= nodes[x][y].x+(int)COW_RADIUS; i++) {
    		    		for (int j = nodes[x][y].y-(int)COW_RADIUS; j <= nodes[x][y].y+(int)COW_RADIUS; j++) {
    		    			if (i == nodes[x][y].x && j == nodes[x][y].y)
    		    				nodes[x][y].punishment = COW_VALUE;
    		    			else if (i >= 0 && i < nodes.length && j >= 0 && j < nodes[0].length) {
    		    				if (nodes[i][j].punishment < COW_VALUE / nodes[x][y].distance(nodes[i][j]))
    		    					nodes[i][j].punishment = COW_VALUE / nodes[x][y].distance(nodes[i][j]);
    		    			}
    		    		}
    		    	}
    	    	}
    	    	
    	    	else if (nodes[x][y].content == WorldModel.OBSTACLE) { 		    	
    		    	for (int i = nodes[x][y].x-(int)OBSTACLE_RADIUS; i <= nodes[x][y].x+(int)OBSTACLE_RADIUS; i++) {
    		    		for (int j = nodes[x][y].y-(int)OBSTACLE_RADIUS; j <= nodes[x][y].y+(int)OBSTACLE_RADIUS; j++) {
    		    			if (i == nodes[x][y].x && j == nodes[x][y].y)
    		    				nodes[x][y].punishment = OBSTACLE_VALUE;
    		    			else if (i >= 0 && i < nodes.length && j >= 0 && j < nodes[0].length) {
    		    				if (model.hasObject(WorldModel.CLEAN, i, j))
    		    					nodes[i][j].punishment = CLEAN_VALUE;
    		    			}
    		    		}
    		    	}
    	    	}	
    	    	else if (nodes[x][y].content == WorldModel.SWITCH) {
    	    		nodes[x][y].punishment = OBSTACLE_VALUE;
    	    	}
    	    	else if (nodes[x][y].content == WorldModel.CORRAL) {
    	    		nodes[x][y].punishment = CORRAL_VALUE;
    	    	}
    	    	else if (nodes[x][y].content == WorldModel.AGENT) {
    	    		nodes[x][y].punishment = AGENT_VALUE;
    	    	}
    	    	else if (nodes[x][y].content == WorldModel.FENCE) {
    	    		nodes[x][y].punishment = FENCE_VALUE;
    	    	} else if (model.agentVisits[x][y] == 0) {
    	    		nodes[x][y].punishment = UNEXPLORED_VALUE;
    	    	}
    			
    		}
    	}
    	
    }
    
    
    /*
        case west     : return new Location(pos.x-1,pos.y);
        case east     : return new Location(pos.x+1,pos.y);
        case north    : return new Location(pos.x,pos.y-1);
        case northeast: return new Location(pos.x+1,pos.y-1);
        case northwest: return new Location(pos.x-1,pos.y-1);
        case south    : return new Location(pos.x,pos.y+1);
        case southeast: return new Location(pos.x+1,pos.y+1);
        case southwest: return new Location(pos.x-1,pos.y+1);
     */
    public String getMove(Node from, Node to) {
        if (to.x < from.x && to.y == from.y)
            return "west";
        else if (to.x > from.x && to.y == from.y)
            return "east";
        else if (to.x == from.x && to.y < from.y)
            return "north";
        else if (to.x > from.x && to.y < from.y)
            return "northeast";
        else if (to.x < from.x && to.y < from.y)
            return "northwest";
        else if (to.x == from.x && to.y > from.y)
            return "south";
        else if (to.x > from.x && to.y > from.y)
            return "southeast";
        else if (to.x < from.x && to.y > from.y)
            return "southwest";
        else
            return "skip";
    }
    
    
    public int getContent(int x, int y) {
    	if (model.hasObject(WorldModel.CLEAN, x, y))
    		return WorldModel.CLEAN;
    	else if (model.hasObject(WorldModel.AGENT, x, y))
    		return WorldModel.AGENT;
    	else if (model.hasObject(WorldModel.OBSTACLE, x, y))
    		return WorldModel.OBSTACLE;
    	else if (model.hasObject(WorldModel.COW, x, y))
    		return WorldModel.COW;
    	else if (model.hasObject(WorldModel.CORRAL, x, y))
    		return WorldModel.CORRAL;
    	else if (model.hasObject(WorldModel.ENEMY, x, y))
    		return WorldModel.ENEMY;
    	else if (model.hasObject(WorldModel.ENEMYCORRAL, x, y))
    		return WorldModel.ENEMYCORRAL;
    	else if (model.hasObject(WorldModel.FENCE, x, y))
    		return WorldModel.FENCE;
    	else if (model.hasObject(WorldModel.SWITCH, x, y))
    		return WorldModel.SWITCH;
    	
    	return 0;
    }
    
    
} // astar

class Node implements Comparable<Node>{
	int x,y;
	int content;
	double walkDistance; // g_score
	double heuristic; // h_score
	double totalDitance; // f_score
	Node from = this;
	double punishment;
	
	public Node(int x, int y) {
		this.x = x;
		this.y = y;
		this.content = 0;
		this.walkDistance = 0;
		this.heuristic = 0.0;
		this.totalDitance = 0.0;
		this.punishment = 0.0;
	}
	
	public int compareTo(Node n) {
		if (this.totalDitance < n.totalDitance)
			return -1;
		else if (this.totalDitance == n.totalDitance)
			return 0;
		else 
			return 1;
	}
	
	public double distance(Node n) {
	    return Math.sqrt(Math.pow(this.x - n.x, 2) + Math.pow(this.y - n.y, 2));
	}
	
	public String toString() {
		return "(" + this.x + "," + this.y + ")";
	}
}
